import copy

from ..utils import get_logger_shortcuts
from .stack import BlockingInstructionStack, HandlerStatus, InstructionStack
from .instruction import Instruction


class Workflow:
    def __init__(self,
            transformators: dict,
            storage,
            logger=None,
        ) -> None:
        
        self.transformators = transformators
        self.storage = storage

        self.logger = logger
        self.log = get_logger_shortcuts(logger)
        self.instruction_stack = BlockingInstructionStack(handlers={
            'subframe': self._handle_subframe_call,
            'transformator_call': self._handle_transformator_call,
            'prepare_transform': self._handle_transform_preparation,
            'execute_transform': self._handle_transform_execution,
        }, logger=self.logger)
        self.instruction_stack.info.name = 'Base stack'

    def _find_transformator_by_items(self, search_items: list) -> str:
        found_transformator = None
        for name, tr_obj in self.transformators.items():
            accepted = True
            for item_name in search_items:
                if not tr_obj.contains_dataitem(item_name):
                    accepted = False
                    break
            if accepted:
                assert found_transformator is None, \
                    f"{repr(search_items)} does not define transformator unambiguously"
                found_transformator = name
        
        assert found_transformator is not None, \
            f"Cannot find transformator containing dataitems {repr(search_items)}"
        return found_transformator

    def execute(self, commands: list, forward: dict={}):
        assert len(self.instruction_stack) == 0, 'Command stack is not empty'
        self.instruction_stack.include_instructions(commands)

        status: HandlerStatus = self.instruction_stack.complete_workflow(forward=forward)
        assert status == HandlerStatus.DONE
        self.log.info("Execution finished normally")
    
    @staticmethod
    def _merge_forwards(instr: Instruction, forward: dict):
        # Priority: instr.info['forward'] > _complete_workflow(forward)
        for key, value in forward.items():
            if key not in instr.info.forward:
                instr.info.forward[key] = value

    def _handle_subframe_call(self, instr: InstructionStack, stack: InstructionStack, forward: dict={}):
        assert instr.is_subframe
        
        self._merge_forwards(instr=instr, forward=forward)

        if instr.is_blocking:
            status: HandlerStatus = instr.complete_workflow(forward=forward)
        else:
            status: HandlerStatus = instr.complete_workflow(hold=stack.is_blocking, forward=forward)
        return status

    def _handle_transformator_call(self, instr: Instruction, stack, forward: dict={}) -> HandlerStatus:
        # Find an appropriate transformator if it is not assigned explicitly
        if instr.info.transformator_name is None:
            search_items = [instr.info.target]
            if instr.info.sources is not None:
                search_items += instr.info.sources
            instr.info.transformator_name = self._find_transformator_by_items(search_items)
        else:
            assert instr.info.transformator_name in self.transformators, \
                f"{instr.info.transformator_name} is not registered in workflow"
        self.log.info(f"Using '{instr.info.transformator_name}' transformator "
            f"to reach the target '{instr.info.target}'")
        
        self._merge_forwards(instr=instr, forward=forward)
        
        assert stack.is_blocking
        stack_frame = stack.followup_substack(instr, blocking=False)

        self.transformators[instr.info.transformator_name].plan_transformation(
            stack_frame=stack_frame,
            transformator_name=instr.info.transformator_name,
            target=instr.info.target,
            sources=instr.info.sources,
            forward=instr.info.forward
        )
        return HandlerStatus.REPEAT
    
    def _handle_transform_preparation(self, instr: Instruction, stack, forward: dict={}) -> HandlerStatus:
        self._merge_forwards(instr=instr, forward=forward)

        stack_frame = stack.followup_substack(instr, blocking=False)

        self.transformators[instr.info.transformator_name].prepare_transform(
            stack_frame=stack_frame,
            transformator_name=instr.info.transformator_name,
            transform_name=instr.info.transform_name,
            transform_path=instr.info.transform_path,
            forward=instr.info.forward
        )
        return HandlerStatus.LATER # Expand transforms asap, then come back to execute them

    def _handle_transform_execution(self, instr: Instruction, stack, forward: dict={}) -> HandlerStatus:
        self._merge_forwards(instr=instr, forward=forward)

        returned_status: HandlerStatus = self.transformators[instr.info.transformator_name].execute_transform(
            stack_frame=stack,
            instruction=instr,
            transformator_name=instr.info.transformator_name,
            transform_name=instr.info.transform_name,
            unmerged_keys=instr.info.unmerged_keys,
            forward=instr.info.forward
        )

        return returned_status



from chemscripts.mylogging import createLogger

class AnyPathSelector:
    def select(seqs, logger=None, **attrs):
        if len(seqs) > 0:
            if logger is None:
                logger = createLogger("ShortestPath")
            if len(seqs) > 1:
                logger.warning("Choosing arbitrary valid route")
            return [seqs[0]]
        else:
            return []


class ShortestPathSelector:
    def select(seqs, g, ag, logger=None, **attrs):
        if logger is None:
            logger = createLogger("ShortestPath")
        logger.info("ENTER")
        unique_seqs = []
        def is_extention(a, b): # Checks if 'a' is extension of 'b'
            if len(a) <= len(b):
                return False
            for itemA in a:
                if itemA not in b:
                    return False
            return True
            # b_idx = len(b) - 1
            # for itemA in reversed(a[len(a)-len(b):]):
            #     if b[b_idx] != itemA:
            #         return False
            #     b_idx -= 1
            # return True

        # print(repr(seqs))

        min_length = None
        for path in seqs:
            if min_length is None or min_length > len(path):
                min_path = path
                min_length = len(path)
            # logger.info(f"Processing {repr(path)}")
            # delete_idxs = []
            # match = False

            # for trial_i, trial_path in enumerate(unique_seqs):
            #     if is_extention(path, trial_path):
            #         print("MATCH")
            #         match=True
            #     elif is_extention(trial_path, path):
            #         delete_idxs.append(trial_i)
            
            # if not match:
            #     unique_seqs.append(path)
            #     logger.info(f"Added {path}")
            
            # delete_idxs.sort(reverse=True)
            # for idx in delete_idxs:
            #     logger.info(f"Removed {unique_seqs[idx]}")
            #     del unique_seqs[idx]
        return [min_path]


class ChooseSourcesPathSelector:
    def __init__(self, sources, strict=True):
        self.sources = sources
        self.strict = strict
    
    def select(self, seqs, g, ag, logger=None, **attrs):
        if logger is None:
            logger = createLogger("ChosenSources")
        
        chosen_paths = []
        for path in seqs:
            # logger.debug(f"------ Processing path {repr(path)} -------")
            all_ends = ag.get_all_ends(path)
            # logger.debug(f"All ends = {repr(all_ends)}")
            # logger.debug(f"Sources = {repr(self.sources)}")
            if self.strict and set(all_ends) == set(self.sources):
                # logger.debug(f"Route '{repr(path)}' is accepted (strict)")
                chosen_paths.append(path)
            elif not self.strict:
                includes = True
                for item in self.sources:
                    if item not in all_ends:
                        # logger.debug(f"Route '{repr(path)}' is discarded")
                        includes = False
                        break
                if includes:
                    chosen_paths.append(path)
        return chosen_paths


import copy

from .decorators import register_dataitem
from .abstract_data import DataItem, ColumnType
from .keynames_layers import GroundKeynamingLayer, KeynamingLayer


@register_dataitem('object')
class ObjectItem(DataItem):
    def __init__(self, keys: list, **kwargs) -> None:
        super().__init__(index_expectation=ColumnType.VARIABLE, **kwargs)

        self.keys = keys
        self._key_mapping = GroundKeynamingLayer(keys=keys)

        self.objects = []

        self.modtime_control = 'manual'

    def include_element(self, object, **keys) -> None:
        """Include one object into `self`

        Args:
            object (Any): object to include
            **kwargs (str): values of all keys
        """

        for key in keys.keys():
            assert key in self.public_keys
        keys = self._key_mapping.map_to_internal(keys)

        if not self._conforms_keys_control(keys):
            return

        if self._is_null:
            self.objects = []
        
        if self.contains_keys(keys=keys):
            index = self._get_index(keys=keys)
            self.objects[index] = object
        else:
            index = len(self.objects)
            self.objects.append(object)

        super().include_element(index=index, keys=keys)
        
    def access_element(self, **keys):
        """Access object by its full set of keys

        Returns:
            Any: the retrieved object

        Raises:
            AssertionError: on incomplete set of provided keys or multiple objects matched
        """

        for key in keys.keys():
            assert key in self.public_keys
        keys = self._key_mapping.map_to_internal(keys)
        index = self._get_index(keys)
        return self.objects[index]

    def get_restricted(self, keys: dict, **kwargs):
        """Get the restricted (filtered) version of the ObjectItem.
        It's very important to pass only the necessary objects into transforms.
        Restricted version can be updated from inside transforms and then manually merged into the main ObjectItem.

        Args:
            keys (Dict[str, Any]): values of keys used for restriction

        Returns:
            ObjectItem: restricted version of ObjectItem.
            It is detached from the `self` ObjectItem, however, the objects themselves are not copied.
        """

        # This takes care of 99% of stuff
        restricted_item = super().get_restricted(keys,
            constructor_kwargs={
                # Pass to the constructor of new ObjectItem
                # a list of all keys that were not restricted
                'keys': [
                    key for key in self._key_names
                    if key not in keys
                ],
                **kwargs
            }
        )

        # On the ObjectItem level, we just need to load the objects into the new ObjectItem,
        # and assign a more natural indexing (0, 1, 2, etc.)
        indices = restricted_item._get_indices()
        index_mapping = {}
        for index in indices:
            new_index = len(restricted_item.objects)
            index_mapping[index] = new_index
            restricted_item.objects.append(self.objects[index])
        restricted_item._rename_indices(index_mapping)

        return restricted_item
    
    def __iadd__(self, other):
        """Merge with another ObjectItem

        Args:
            other (ObjectItem): Usually, restricted version of ObjectItem
            produced by transform is merged into the main ObjectItem

        Returns:
            ObjectItem: updated main ObjectItem
        """

        # This only performs a few assertions
        super().__iadd__(other)

        for keys, object in other.private_iterator():
            self.include_element(object=object, **keys)
        return self
    
    def restore_key_mapping(self) -> None:
        self._key_mapping = {
            keyname: keyname
            for keyname in self.keys
        }

    def __iter__(self):
        pub_keys = self.public_keys
        return self._base_iter(
            call=lambda index, keys: (
                self.objects[index],
                {
                    key: value
                    for key, value in self._key_mapping.map_to_external(keys).items()
                    if key in pub_keys
                }
            )
        )
    
    def private_iterator(self):
        return self._base_iter(
            call=lambda index, keys: (keys, self.objects[index])
        )

    def cleanup(self) -> None:
        # Delete all stored object refs
        self.objects = []

        # This will cleanup the table
        super().cleanup()

    def __contains__(self, item):
        return self.contains_keys(item['keys'])

    @property
    def _key_names(self):
        """Access the list of keys of this ObjectItem

        Returns:
            List[str]: List of keys
        """
        return copy.copy(self.keys) + super()._key_names
    
    @property
    def _is_null(self):
        return self.objects is None



import networkx as nx
import pandas as pd

from ..utils import get_logger_shortcuts

class AccessGraph:
    def __init__(self, inp_g, node_set, logger=None):
        self.log = get_logger_shortcuts(logger)

        self.g = nx.DiGraph()
        self.g.add_nodes_from(node_set)
        for nodeA, nodeB in inp_g.edges:
            if nodeA in node_set and nodeB in node_set:
                self.g.add_edge(nodeA, nodeB)
        self.log.debug(f"AccessGraph nodes = {repr(list(self.g.nodes))}")
        self.log.debug(f"AccessGraph edges = {repr(list(self.g.edges))}")
        
        for node in self.g.nodes:
            self.g.nodes[node]['is_item'] = inp_g.nodes[node]['is_item']
            if self.g.nodes[node]['is_item']:
                if 'item' in inp_g.nodes[node] and inp_g.nodes[node]['item'].non_empty:
                    self.g.nodes[node]['known'] = True
                    self.g.nodes[node]['active'] = True
                    self.log.debug(f"Node '{node}' is known")
                else:
                    self.g.nodes[node]['known'] = False
                    self.g.nodes[node]['active'] = False
                    self.log.debug(f"Node '{node}' is not known")
            else:
                self.g.nodes[node]['active'] = False
        
        self.log.debug("Before propagation:\n" + self.get_state())
        converged = False
        while not converged:
            converged = True
            for node in self.g.nodes:
                if not self.g.nodes[node]['is_item']:
                    all_inputs_active = True
                    for nb in self.g.predecessors(node):
                        all_inputs_active = all_inputs_active and self.g.nodes[nb]['active']
                    
                    if not self.g.nodes[node]['active'] and all_inputs_active:
                        self.g.nodes[node]['active'] = True
                        converged = False
                    elif self.g.nodes[node]['active'] and not all_inputs_active:
                        self.g.nodes[node]['active'] = False
                        converged = False
                
            for node in self.g.nodes:
                if self.g.nodes[node]['is_item']:
                    some_source_is_active = False
                    for nb in self.g.predecessors(node):
                        some_source_is_active = some_source_is_active or self.g.nodes[nb]['active']

                    if not self.g.nodes[node]['active'] and some_source_is_active:
                        self.g.nodes[node]['active'] = True
                        converged = False
        self.log.debug("After propagation:\n" + self.get_state())
    
    def paths_iterate(self, target):
        prospects = [[node] for node in self.g.predecessors(target) if self.g.nodes[node]['active']]
        while len(prospects) > 0:
            cur_prospect = prospects.pop()
            loose_ends = []
            all_ends = []

            incorrect_prospect = False
            for node in self.g.nodes:
                required = False
                employing_tr_idxs = []
                for employing_transform in self.g.successors(node):
                    if employing_transform in cur_prospect:
                        employing_tr_idxs.append(cur_prospect.index(employing_transform))
                        required = True
                
                if not required:
                    continue
                
                n_formed = 0
                forming_idxs = []
                for forming_transform in self.g.predecessors(node):
                    if forming_transform in cur_prospect:
                        n_formed += 1
                        forming_idxs.append(cur_prospect.index(forming_transform))
                if n_formed > 1:
                    incorrect_prospect = True
                elif n_formed == 1:
                    forming_idx = forming_idxs[0]
                    for employing_tr_idx in employing_tr_idxs:
                        if employing_tr_idx > forming_idx:
                            incorrect_prospect = True
                            break
                else:
                    all_ends.append(node)
                    if not self.g.nodes[node]['known']:
                        loose_ends.append(node)
                
                if incorrect_prospect:
                    break
            
            if incorrect_prospect:
                cur_prospect.reverse()
                self.log.warning(f"Incorrect prospect was encountered and ignored: {repr(cur_prospect)}")
                continue 

            if len(loose_ends) == 0: # Yield a final and checked sequence
                ret_list = list(reversed(cur_prospect))
                yield ret_list
            
            # Look for prospects if there are loose ends
            for item_node in all_ends:
                for next_transorm in self.g.predecessors(item_node):
                    if not self.g.nodes[next_transorm]['active']:
                    #     or next_transorm in cur_prospect:
                        continue
                    new_prospect = [*cur_prospect, next_transorm]
                    if new_prospect not in prospects:
                        prospects.append(new_prospect)
        
    def get_state(self):
        data = []
        for node in self.g.nodes:
            data.append({**{'node': node}, **{key: value for key, value in self.g.nodes[node].items()}})
        df = pd.DataFrame(data)
        return repr(df[df['is_item']])
    
    def get_all_ends(self, seq):
        seq = list(reversed(seq))
        all_ends = []

        for node in self.g.nodes:
            required = False
            employing_tr_idxs = []
            for employing_transform in self.g.successors(node):
                if employing_transform in seq:
                    employing_tr_idxs.append(seq.index(employing_transform))
                    required = True
            
            if not required:
                continue
            
            n_formed = 0
            forming_idxs = []
            for forming_transform in self.g.predecessors(node):
                if forming_transform in seq:
                    n_formed += 1
                    forming_idxs.append(seq.index(forming_transform))
            
            # Skipped some checks here
            
            if n_formed == 0:
                all_ends.append(node)
        return all_ends



import copy
import inspect

from ..utils import get_logger_shortcuts
from ..transforms import Transform


class TransformConcrete:
    def __init__(self,
            transformator,
            transform_name: str,
            logger=None
        ) -> None:
        
        self.transformator = transformator
        self.transform_name = transform_name

        self.logger = logger
        self.log = get_logger_shortcuts(logger)
    
    def prepare_instance(self, forward: dict={}) -> None:
            
        self.log.info(f"Preparing for '{self.transform_name}' transform")
        self.cur_transform: Transform = self.transformator.transformations[self.transform_name]

        self.args_list = {
            method_name: inspect.getfullargspec(method)[0]
            for method_name, method in self.cur_transform.methods.items()
        }
        assert 'exec' in self.args_list, \
            f"Entry point ('exec'-method) is not implemented for transform '{self.transform_name}'"
        
        aware_keys = self.cur_transform.AWARE_KEYS
        merged_keys = self.cur_transform.MERGED_KEYS
        assert len(set(aware_keys).intersection(merged_keys)) == 0, \
            f"Aware and Merge sets overlap. Aware={repr(aware_keys)}, Merged={repr(merged_keys)}"

        # Construct lists of keys
        self.keys_lists = {}
        self.source_keys = []
        for source_name in self.cur_transform.SOURCE_ITEMS:
            self.keys_lists[source_name] = self.transformator.graph.nodes[source_name]['item'].public_keys
            for key in self.keys_lists[source_name]:
                if key not in self.source_keys:
                    self.source_keys.append(key)
        for cur_target_name in self.cur_transform.TARGET_ITEMS:
            self.keys_lists[cur_target_name] = self.transformator.graph.nodes[cur_target_name]['item'].public_keys
        for note_name in self.cur_transform.NOTE_ITEMS:
            self.keys_lists[note_name] = self.transformator.graph.nodes[note_name]['item'].public_keys
        self.log.debug("keys_lists = " + repr(self.keys_lists))

        # Check that requested keys are present
        aware_okay = True
        for key in aware_keys:
            if key not in self.source_keys:
                aware_okay = False
                self.log.error(f"Requested aware key '{key}' is not involved in transform '{self.transform_name}'")
        merged_okay = True
        for key in merged_keys:
            if key not in self.source_keys:
                merged_okay = False
                self.log.error(f"Requested merged key '{key}' is not involved in transform '{self.transform_name}'")
        assert aware_okay and merged_okay, f"Failed key check on the stage '{self.transform_name}'"
        
        # Check for overlapping naming of keys, items, forwarded args, etc.
        itemnames_set = set()
        keys_set = set()
        forward_stuff = list(forward.keys())
        for l in self.keys_lists.values():
            keys_set.update(set(l))
        for l in [self.cur_transform.SOURCE_ITEMS, self.cur_transform.TARGET_ITEMS, self.cur_transform.NOTE_ITEMS]:
            itemnames_set.update(set(l))

        def check_overlap(**attrs):
            assert len(attrs) == 2
            names = list(attrs.keys())
            sets = list(attrs.values())
            intersection = sets[0].intersection(sets[1])
            assert len(intersection) == 0, f"{names[0]} and {names[1]} share element(s): {repr(intersection)}"
        check_overlap(item_names=itemnames_set, key_names=keys_set)
        check_overlap(item_names=itemnames_set, forwarded_keys=forward_stuff)
        check_overlap(key_names=keys_set, forwarded_keys=forward_stuff)

        # Check prerequisites for unaware, aware and merged keys
        for key in self.source_keys:
            each_source = True
            each_target = True
            one_source = False
            one_target = False # "one_target and not one_source" is a criteria for new key
            for source_name in self.cur_transform.SOURCE_ITEMS:
                if key in self.keys_lists[source_name]:
                    one_source = True
                else:
                    each_source = False
            for cur_target_name in self.cur_transform.TARGET_ITEMS:
                if key in self.keys_lists[cur_target_name]:
                    one_target = True
                else:
                    each_target = False
            
            if key in aware_keys: # aware
                assert each_target and one_source, f"Necessary condition for aware key is not satisfied for '{key}'"
                self.log.debug(f"Key '{key}' is in aware mode")
            elif key in merged_keys: # merged
                assert one_source, f"Necessary condition for merged key is not satisfied for '{key}'"
                self.log.debug(f"Key '{key}' is in merged mode")
            else: # unaware
                assert each_source and each_target, f"Necessary condition for unaware key is not satisfied for '{key}'"
                self.log.debug(f"Key '{key}' is in unaware mode")

        self.nonmerged_keys = copy.copy(self.source_keys)
        for key in merged_keys:
            del self.nonmerged_keys[self.nonmerged_keys.index(key)]
        if self.transform_name == 'get_nbo_cube':
            print('HERE')

    def get_key_combinations(self):
        return self.transformator.storage.key_combinations(
            items=self.cur_transform.SOURCE_ITEMS,
            nonmerged_keys=self.nonmerged_keys
        )



import inspect
import copy
import functools
import itertools

from ..dataitems import DataItem


class Transform:
    DEFAULTS = { # TODO This part really lacks descriptive comments on wtf is this
        'NAME': 'UNTITLED',
        'SOURCE_ITEMS': [],
        'TARGET_ITEMS': [],
        'NOTE_ITEMS': [],
        'AWARE_KEYS': [],
        'AWARE_ENFORCED_KEYS': [],
        'MERGED_KEYS': [],
        'GREEDY_TARGETS': [],

        'IS_PROXY': False,
    }

    def __init__(self, **attrs):
        self.specs = {}
        self.methods = {}

        for key, value in attrs.items():
            self._add_new_spec(key, value)
        for key, value in Transform.DEFAULTS.items():
            if key not in attrs:
                self._add_new_spec(key, copy.copy(value))
        
        self.specific_args = {
            'self': self,
        }
    
    def set_method(self, arg):
        """Specify method(s) for the transform

        Args:
            arg (Callable or Dict[str, Callable]): Function or Dict of functions to be included as methods

        Raises:
            AssertionError: Failed typechecks or unexpected function name
            ValueError: arg of unexpected type
        """
        new_methods = {}

        if callable(arg):
            formatting_error = 'Expected the method to have format "{Name}_{Method}"'
            if arg.__name__.count('_') == 1:
                if self.NAME != self.DEFAULTS['NAME']:
                    assert arg.__name__.split('_')[0] == self.NAME, formatting_error
                new_methods[arg.__name__.split('_')[1]] = arg
            elif arg.__name__.count('_') == 0:
                new_methods[arg.__name__] = arg
            else:
                raise AssertionError(formatting_error)
        elif isinstance(arg, dict):
            for name, f in arg.items():
                assert isinstance(name, str), f'{repr(name)} is not a string'
                assert callable(f), f'{repr(f)} is not callable'
            new_methods = arg
        else:
            raise ValueError(f"Cannot set {repr(arg)} as method")
        
        for name, f in new_methods.items():
            self._add_new_method(name, f)
        return self
    
    def extend(self, base_transform, keys_mapping={}, inherit_methods=None):
        """Implements inheritance for transforms

        Args:
            base_transform (Transform): The parent transform
            keys_mapping (str/dict, optional): Mapping from old keys to new in the form
            `"OldA->NewA OldB->NewB"`. Defaults to "".
            inherit_methods (List[str], optional): Names of methods to be inherited. Defaults to None - inherit all methods.
        """

        # Basic processing of the base transform 
        assert 'base' not in self.specific_args, f"'{self.NAME}' already inherits from '{self.specific_args['base'].NAME}'"
        assert isinstance(base_transform, Transform), f"'{self.NAME}' transform can extend only transform objects"

        # Forwarding of methods from the base transform
        # (1) if they are NOT declared within self: create proxy and access as self.f(...)
        # (2) if they ARE declared within self: create proxy and access as self.base.f(...)
        
        # each key mapping renames either names of '*_ITEMS' or '*_KEYS'
        if isinstance(keys_mapping, dict):
            key_change = keys_mapping
        elif isinstance(keys_mapping, str):
            keys_mapping = '\n'.join([
                line
                for line in keys_mapping.splitlines()
                if not line.lstrip().startswith('#') # Ignore comment lines
            ])
            key_change = {}
            for part in keys_mapping.split():
                sides = part.split('->')
                key_change[sides[1]] = sides[0] # Maps 'parent_key' -> 'child_key'
        else:
            raise AssertionError('Unexpected type of key mapping')
        
        # Maps 'child_key' -> 'parent_key' (internal -> external)
        key_change_inverse = {value: key for key, value in key_change.items()}

        self.specific_args['base'] = base_transform._get_transform_proxy(key_change)

        def assign_proxy_method(method_name, unused_parent_keys):
            method = getattr(base_transform, method_name)
            assert callable(method), f'{repr(method)} is not a callable method prototype'
            
            specific_args = {}
            args_list = inspect.getfullargspec(method)[0]
            if 'self' in args_list:
                specific_args['self'] = self.base
            if 'base' in args_list:
                specific_args['base'] = base_transform.base
            
            corrected_argnames = [
                argname
                for argname in args_list
                if argname in key_change
            ]
            for argname in corrected_argnames:
                if argname in unused_parent_keys:
                    del unused_parent_keys[unused_parent_keys.index(argname)]

            keep_argnames = [
                argname
                for argname in args_list
                if argname not in specific_args and argname not in corrected_argnames
            ]
            new_arglist = [
                *[
                    key_change[arg]
                    for arg in corrected_argnames
                ],
                *keep_argnames,
            ]

            # This decorator is the sugar to keep metadata of the parent method
            @functools.wraps(method)
            def proxy_method(**args):
                corrected_args = {
                    old_argname: args[key_change[old_argname]]
                    for old_argname in corrected_argnames
                }
                kept_args = {
                    argname: args[argname]
                    for argname in keep_argnames
                }
                complete_args = {**corrected_args, **kept_args, **specific_args}

                for argobject in complete_args.values():
                    if isinstance(argobject, DataItem):
                        argobject.with_renamed_keys(key_change_inverse)
                
                # Execution itself
                result = method(**complete_args)

                for argobject in complete_args.values():
                    if isinstance(argobject, DataItem):
                        argobject.release_keyname_layer()
                
                return result

            new_params = [
                inspect.Parameter(argname, inspect.Parameter.POSITIONAL_OR_KEYWORD)
                for argname in new_arglist
            ]

            proxy_method.__signature__ = inspect.signature(proxy_method).replace(parameters=new_params)

            self.base._add_new_method(method_name, proxy_method)
            if method_name not in self.methods:
                self._add_new_method(method_name, proxy_method)
            if self.specs['NAME'] == 'get_nbo_cube':
                print('HI')

        if inherit_methods is None:
            methods_iter = base_transform.methods.keys()
        else:
            for method_name in inherit_methods:
                assert method_name in base_transform.methods, \
                    f"Requested method '{method_name}' is not present in the parent transform '{base_transform.NAME}'"
            methods_iter = inherit_methods
        
        unused_parent_keys = list(key_change.keys())
        for method_name in methods_iter:
            assign_proxy_method(method_name, unused_parent_keys)

        unused_maps = {
            key: key_change[key]
            for key in unused_parent_keys
        }
        # Disabled this asserting since it doesn't account for renamed keys of DataItems
        # TODO Make optional format with ==> or something to indicate that this renaming must be checked
        # assert len(unused_maps) == 0, f"Unused key mappings: {repr(unused_maps)}"

    def _add_new_spec(self, spec_name, value):
        """Add specification to this transform

        Args:
            spec_name (str): Name of the new attr. Prefer UPPER_CASE. Can later acces it via self.SPEC_NAME
            f (Callable): Calls will be redirected to this callable object
        """
        assert spec_name == spec_name.upper(), f'Prefer UPPER_CASE transform specs. Got "{repr(spec_name)}"'
        assert spec_name not in self.specs, f'Attempting to reassign spec {spec_name} of transform {self.NAME}'
        self.specs[spec_name] = value
        
    def _add_new_method(self, method_name, f):
        """Add method to this transform.

        Args:
            method_name (str): Name of the method. Later can be called as `self.method_name`
            f (Callable): Calls will be redirected to this callable object
        """

        # In some cases inherited methods must be overwritten by implemented ones, so this assertion is not needed
        # assert method_name not in self.methods, f'Attempting to reassign method {method_name}{inspect.signature(f)} of {proxy_flag}transform {self.NAME}'

        # If '**something' is present in the argument list, we replace it with all missing necessary args
        argspec = inspect.getfullargspec(f)
        f_varkw = argspec.varkw
        if f_varkw is not None: # It is a string (name of the 'something') or None
            old_sig = inspect.signature(f)
            old_sig_paramnames = [
                param.name
                for param in old_sig.parameters.values()
            ]

            new_params = [
                param
                for param in old_sig.parameters.values()
                if param.kind != param.VAR_KEYWORD
            ] + [
                inspect.Parameter(argname, inspect.Parameter.POSITIONAL_OR_KEYWORD)
                for argname in self._minimum_expected_args()
                if argname not in old_sig_paramnames
            ]

            new_sig = old_sig.replace(parameters=new_params)
            f.__signature__ = new_sig

        # This will make error messages much clearer
        if self.IS_PROXY:
            proxy_flag = "proxy-"
        else:
            proxy_flag = ""

        # Check that all necessary args are present
        arg_names = inspect.getfullargspec(f)[0]
        for argname in self._minimum_expected_args():
            assert argname in arg_names, f"Required arg '{argname}' is not supported " \
                f"by method {f.__name__} of the {proxy_flag}transform {self.NAME}"

        self.methods[method_name] = f

    def _has_native_method(self, method_name):
        """Check if transform has this method implemented natively (accessible directly via `self.method_name`)

        Args:
            method_name (str): Name of the method to look for.

        Returns:
            bool: result
        """
        return method_name in self.methods

    def _get_transform_proxy(self, rename_map):
        """Create transform proxy of the base tranform (called from parent transform, obviously)
        so that it can be put in the `child_transform.specific_args['base']`.
        To do this, all the specs having different name in the child transform are renamed here.

        Args:
            rename_map (Dict[str, str/list]): renaming map (can be an empty dict)

        Returns:
            Transform: The proxy for base transform that can be accessed from child transform.
            It does not have any methods at this point, but they will be added in the 'extend'.
        """

        for key, value in self.specs.items():
            if key.endswith('_KEYS') or key.endswith('_ITEMS'):
                assert isinstance(value, list), f'Unexpected type of {key} spec'

        new_specs = {
            'NAME': self.NAME,
            'IS_PROXY': True,
            **{
                key: [
                    rename_map[item] if item in rename_map
                    else item
                    for item in value
                ]
                for key, value in self.specs.items()
                if isinstance(value, list) and (key.endswith('_KEYS') or key.endswith('_ITEMS'))
            },
        }

        result = Transform(**new_specs)
        if 'base' in self.specific_args:
            result.extend(self.base, rename_map)
        return result

    def _minimum_expected_args(self) -> list:
        """
        Get the minumum set of arguments that all methods of this transform must take

        Returns:
            List[str]: list of arguments
        """
        
        res = []
        for key in [
                    'SOURCE_ITEMS',
                    'TARGET_ITEMS',
                    'NOTE_ITEMS',
                    'AWARE_KEYS',
                    'AWARE_ENFORCED_KEYS',
                ]:
            res += self.specs[key]
        return res
        # return list(itertools.chain( # This is a 'short' way of list concatenation (:
        #     *(
        #         self.specs[key] for key in [
        #             'SOURCE_ITEMS',
        #             'TARGET_ITEMS',
        #             'NOTE_ITEMS',
        #             'AWARE_KEYS',
        #             'AWARE_ENFORCED_KEYS',
        #         ]
        #     )
        # ))

    def __contains__(self, attr_name):
        if attr_name in self.specs:
            return True
        
        if attr_name in self.methods:
            return True

        if attr_name in self.specific_args:
            return True
        
        if 'base' in self.specific_args and attr_name in self.base:
            return True

        return False

    def greedy_on(self, *args):
        # skip calling transform if at least one target with unmerged keys exists
        if len(args) == 0:
            self.GREEDY_TARGETS.append(set(x for x in self.TARGET_ITEMS))
        elif len(args) == 1 and callable(args[0]):
            self.GREEDY_TARGETS.append(args[0])
        else:
            for arg in args:
                assert arg in self.TARGET_ITEMS, f"'{arg}' is not a target of transform '{self.NAME}'"
            assert set(args) not in self.GREEDY_TARGETS
            self.GREEDY_TARGETS.append(set(args))
        return self

    def __getattr__(self, attr_name):
        if attr_name in self.specs:
            return self.specs[attr_name]
        
        if attr_name in self.methods:
            return self.methods[attr_name]

        if attr_name in self.specific_args:
            return self.specific_args[attr_name]
        
        if 'base' in self and attr_name in self.base:
            return self.base[attr_name]

        raise AttributeError(f"'{self.NAME}' transform does not have an attribute '{attr_name}'")



import copy
import time
from typing import Callable

from enum import Enum

from .instruction import Instruction
from ..utils import get_logger_shortcuts


class HandlerStatus(Enum):
    DONE = 1
    REPEAT = 2
    LATER = 3


class InstructionStack(Instruction):
    # Inheriting Instruction because stacks can also serve as "instructions" being substacks in other stacks

    def __init__(self, data: Instruction=None, handlers: dict=None, parent_stack=None, logger=None):
        self.instructions = []
        self.handlers = handlers
        self.parent_stack = parent_stack
        assert (self.parent_stack is not None) ^ (self.handlers is not None), \
            f"InstructionStack must be initialized with either parent_stack or handlers"

        self.logger = logger
        self.log = get_logger_shortcuts(logger)

        self.current_instruction_idx = None
        self.next_instruction_idx = None

        if data is None:
            super().__init__({})
        else:
            super().__init__(data.info)
            
    def include_instructions(self, instr_list: list) -> None:
        for instr in reversed(instr_list):
            self.include_instruction(instr)
            
    def include_instruction(self, instruction: Instruction) -> None:
        assert isinstance(instruction, Instruction), \
            f"Attempting to include '{repr(instruction)}' of non-Instruction type"
        
        if 'name' in instruction.info:
            for item in self.instructions:
                if item.info['name'] == instruction.info['name']:
                    self.log.warning("Instruction with the name "
                        f"'{item.info['name']}' already registered in non-blocking stack")
        
        self.instructions.append(instruction)

        self.log.info(f"Stack {repr(self.info.name)}: Included instruction {repr(instruction.info.name)}")

    def __len__(self) -> int:
        return len(self.instructions)

    def followup_instruction(self, instruction: Instruction):
        assert self.current_instruction_idx is not None, \
            f"Cannot follow-up instruction {repr(instruction.info)}, when stack execution is not running"
        
        self.instructions[self.current_instruction_idx] = instruction

    def followup_substack(self, instruction: Instruction, blocking: bool=True):
        assert isinstance(instruction, Instruction)

        new_element = Instruction(data=copy.copy(instruction.info))
        new_element.info['type'] = 'subframe'

        if blocking:
            new_element = BlockingInstructionStack(new_element, parent_stack=self, logger=self.logger)
        else:
            new_element = NonBlockingInstructionStack(new_element, parent_stack=self, logger=self.logger)

        self.followup_instruction(new_element)
        return new_element
    
    def _get_handlers(self):
        if self.handlers is None:
            assert self.parent_stack is not None, "Both parent stack and instruction handlers are missing"
            return self.parent_stack._get_handlers()
        else:
            return self.handlers

    def _get_handler_for(self, instruction_type: str) -> Callable:
        handlers = self._get_handlers()
        assert instruction_type in handlers, \
            f"No handler for instruction of type '{instruction_type}'"
        return handlers[instruction_type]
    
    def _repr_instruction(self, instruction: Instruction, actual_index:int, show_index: int) -> list:
        if actual_index == self.current_instruction_idx and instruction.is_subframe:
            cur_flag = ' > '
        elif actual_index == self.current_instruction_idx and not instruction.is_subframe:
            cur_flag = '—> '
        else:
            cur_flag = '   '

        name = "<UNTITLED>"
        if 'name' in instruction.info:
            name = instruction.info.name

        if instruction.is_subframe:
            return [f"{cur_flag}{show_index+1}) {name}"] + [
                '  ' + line
                for line in instruction._repr_as_lines()
            ]
        else:
            return [f"{cur_flag}{show_index+1}. {name}"]

    def _repr_as_lines(self) -> list:
        lines = []
        show_index = 0
        for index in range(len(self.instructions) - 1, -1, -1):
            instruction = self.instructions[index]
            lines += self._repr_instruction(instruction, actual_index=index, show_index=show_index)
            show_index += 1

        return lines
    
    def print_stack(self, stream=None) -> None:
        newline = '\n'
        text = f"""
===============================
WORKFLOW STACK:
{newline.join(self._repr_as_lines())}
===============================

1) substack
1. task
 > active stack frame
—> active task
"""
        if stream is None:
            self.log.info(text)
        else:
            stream.write(text)
    
    @property
    def base_stack(self):
        if self.parent_stack is not None:
            return self.parent_stack.base_stack
        else:
            return self


class BlockingInstructionStack(InstructionStack):
    def complete_workflow(self, forward: dict={}) -> None:
        while len(self) > 0:

            if self.next_instruction_idx is not None:
                self.current_instruction_idx = self.next_instruction_idx
                self.next_instruction_idx = None
            else:
                self.current_instruction_idx = len(self) - 1

            current_instruction = self.instructions[self.current_instruction_idx]
            if current_instruction.is_active:
                handling_method = self._get_handler_for(current_instruction.info.type)
                self.log.info(f"Stack {repr(self.info.name)} (blocking): Running type={repr(current_instruction.info.type)} name={repr(current_instruction.info.name)}")
                status: HandlerStatus = handling_method(instr=current_instruction, stack=self, forward=forward)
            else:
                raise NotImplementedError
            
            if status == HandlerStatus.DONE:
                self.instructions[self.current_instruction_idx].activate_dependents(stack=self)
                del self.instructions[self.current_instruction_idx]
            elif status == HandlerStatus.REPEAT:
                self.next_instruction_idx = self.current_instruction_idx
            else:
                raise RuntimeError("BlockingInstructionStack cannot process "
                    f"the returned HandlerStatus='{repr(status)}'")
        
        self.current_instruction_idx = None
        self.next_instruction_idx = None

        return HandlerStatus.DONE

    @property
    def is_blocking(self):
        return True


class NonBlockingInstructionStack(InstructionStack):
    def complete_workflow(self, hold: bool, forward: dict={}) -> HandlerStatus:
        if hold:
            while len(self) > 0:
                self._do_single_round(forward=forward)
                time.sleep(1)
        else:
            self._do_single_round(forward=forward)

        if len(self) == 0:
            return HandlerStatus.DONE
        else:
            return HandlerStatus.LATER # Can happen only when hold=False

    def include_instruction(self, instruction: Instruction) -> None:
        assert 'name' in instruction.info, \
            f"'{repr(instruction.info)}' must contain 'name' when it's included in non-blocking stack"
        
        for item in self.instructions:
            assert item.info['name'] != instruction.info['name'], \
                f"Instruction with the name '{item.info['name']}' already registered in non-blocking stack"
            assert 'activated_by' in instruction.info and 'activates' in instruction.info, \
                    "Instructions in nonblocking stack must contain 'activated_by' and 'activates' attrs"
        super().include_instruction(instruction)

    def _do_single_round(self, forward: dict={}) -> None:
        self.next_instruction_idx = len(self) - 1
        while self.next_instruction_idx != -1:

            self.current_instruction_idx = self.next_instruction_idx
            current_instruction = self.instructions[self.current_instruction_idx]
            if current_instruction.is_active:
                if 'stack_stream' in forward and forward['stack_stream'] is not None:
                    self.base_stack.print_stack(forward['stack_stream'])
                assert 'activated_by' in current_instruction.info and 'activates' in current_instruction.info, \
                    "Instructions in nonblocking stack must contain 'activated_by' and 'activates' attrs"
                handling_method = self._get_handler_for(current_instruction.info.type)
                self.log.info(f"Stack {repr(self.info.name)} (nonblocking): Running type={repr(current_instruction.info.type)} name={repr(current_instruction.info.name)}")
                status: HandlerStatus = handling_method(instr=current_instruction, stack=self, forward=forward)
                time.sleep(0.01) # This is for visual effect when live stack is printed
            else:
                status = HandlerStatus.LATER

            if status == HandlerStatus.DONE:
                # Remove instruction if we don't have to come back to it
                self.instructions[self.current_instruction_idx].activate_dependents(stack=self)
                del self.instructions[self.current_instruction_idx]
                self.next_instruction_idx = self.current_instruction_idx - 1
            elif status == HandlerStatus.REPEAT:
                self.next_instruction_idx = self.current_instruction_idx
            elif status == HandlerStatus.LATER:
                self.next_instruction_idx = self.current_instruction_idx - 1
            else:
                raise RuntimeError("NonBlockingInstructionStack cannot process "
                    f"the returned HandlerStatus='{repr(status)}'")
            
        self.current_instruction_idx = None
        self.next_instruction_idx = None

    @property
    def is_blocking(self):
        return False



from .decorators import REGISTERED_DATAITEM_TYPES
from .abstract_data import DataItem, ColumnType
from .object import ObjectItem
from .abstract_path import PathItem
from .file import FileItem
from .directories import DirectoryItem



import pathlib
from setuptools import setup, find_packages

PACKAGE_NAME = 'pysquared'

CWD = pathlib.Path(__file__).parent

README_PATH = (CWD / "README.md").read_text()

setup(
    name=PACKAGE_NAME,
    version="0.0.1",
    description="Pipeline automation framework for Python",
    long_description=README_PATH,
    long_description_content_type="text/markdown",
    author="Nikolai Krivoshchapov",
    author_email="nikolai.krivoshchapov@gmail.com",
    license="MIT",
    packages=find_packages(),
    install_requires=[
        'numpy',
        'networkx',
        'pandas',
        'textwrap3',
        'prompt-toolkit',
        'openpyxl',
        'xlcalculator',
    ]
)



from .transform_concrete import TransformConcrete
from ..utils import get_logger_shortcuts

from .transform_utils import TransformState, TransformResult


class TransformExecutor:
    def __init__(self,
            transform_instance: TransformConcrete,
            transformator_name: str,
            transform_name: str,
            logger=None
        ) -> None:

        self.transform_instance = transform_instance
        self.transformator = transform_instance.transformator
        self.abstract_transform = self.transform_instance.cur_transform
        
        # Be default, the 'exec' method serves as entry point for all transforms
        self.next_method = 'exec'
        self.next_internals = {}

        self.transformator_name = transformator_name
        self.transform_name = transform_name

        self.logger = logger
        self.log = get_logger_shortcuts(logger)

        self.initialized_for = None

    def initialize_restrictions(self, unmerged_keys: dict, keys_control=None) -> None:
        keys_lists = self.transform_instance.keys_lists
        self.item_restrictions = {}
        for item_list in (self.abstract_transform.SOURCE_ITEMS, 
                          self.abstract_transform.TARGET_ITEMS,
                          self.abstract_transform.NOTE_ITEMS):
            for item_name in item_list:
                self.item_restrictions[item_name] = self.transformator.graph.nodes[item_name]['item'].get_restricted({
                    key: value for key, value in unmerged_keys.items() if key in keys_lists[item_name]
                }, keys_control=keys_control)
        
        skip = False
        for item_name in self.abstract_transform.SOURCE_ITEMS:
            if self.item_restrictions[item_name].is_empty:
                self.log.error(f"Item '{item_name}' is empty for keys '{repr(unmerged_keys)}'")
                skip = True
        if skip:
            raise RuntimeError('SKIP!?')

    def initialize_for_method(self, method_name: str, unmerged_keys: dict, forward: dict={}) -> None:
        args_list = self.transform_instance.args_list[method_name]

        aware_data = {
            key: value
            for key, value in unmerged_keys.items()
            if key in self.abstract_transform.AWARE_KEYS
        }

        forwarded_args = {
            key: value
            for key, value in forward.items()
            if key in args_list
        }
        
        specific_args = {
            key: value
            for key, value in self.abstract_transform.specific_args.items()
            if key in args_list
        }

        self.execution_args = {
            **self.item_restrictions, # Restricted items are universal for all method calls
            **aware_data, # Are also universal for all method calls
            **forwarded_args, # Passed if required
            **specific_args # Args that are supported by this abstract transform (self and base)
        }

    def execute(self, unmerged_keys: dict, forward: dict={}) -> TransformResult:
        # Only for the first execution
        if not hasattr(self, 'item_restrictions'):
            keys_control = None
            if 'keys_control' in forward:
                keys_control = forward['keys_control']
            self.initialize_restrictions(unmerged_keys, keys_control)
        
        if 'GREEDY_TARGETS' in self.abstract_transform.specs:
            for greedy_targets in self.abstract_transform.GREEDY_TARGETS:
                if callable(greedy_targets):
                    skip_allowed = greedy_targets({
                        key: value
                        for key, value in self.item_restrictions.items()
                        if key in self.abstract_transform.TARGET_ITEMS
                    })
                else: # It's a list of str then ...
                    assert len(greedy_targets) > 0, f"Set of greedy targets is present but it's empty"
                    skip_allowed = True
                    for target_item in greedy_targets:
                        skip_allowed = skip_allowed and (len(self.item_restrictions[target_item]) > 0)
                if skip_allowed:
                    return TransformResult.FINISHED

        if self.initialized_for != self.next_method:
            self.initialize_for_method(
                method_name=self.next_method,
                unmerged_keys=unmerged_keys,
                forward=forward
            )

        # Execute method and convert the returned value to TransformState instance
        self.log.info(f"Executing the method '{self.next_method}' of transform '{self.abstract_transform.specs['NAME']}'")
        state: TransformState = TransformState.returnvalue_to_state(
            self.abstract_transform.methods[self.next_method](**self.execution_args, **self.next_internals)
        )

        if state.result == TransformResult.FINISHED:
            for cur_target_name in self.abstract_transform.TARGET_ITEMS:
                if self.item_restrictions[cur_target_name].non_empty:
                    self.transformator.graph.nodes[cur_target_name]['item'] += self.item_restrictions[cur_target_name]
                else:
                    self.log.warning(f"Target '{cur_target_name}' is empty after executing '{self.abstract_transform.NAME}'. " \
                        f"Restriction={repr(unmerged_keys)}")
            for restricted_item in self.item_restrictions.values():
                restricted_item.retire()
            self.log.info(f"Step {self.transform_name} finished successfully for {repr(unmerged_keys)}")
        elif state.result == TransformResult.FAILED:
            for restricted_item in self.item_restrictions.values():
                restricted_item.retire()
        else: # either REPEAT or LATER
            self.next_method = state.comeback_method
            self.next_internals = state.internals

        return state.result



import networkx as nx
import itertools
import inspect
import time

from chemscripts.mylogging import createLogger
from .accessgraph import AccessGraph
from .transform_concrete import TransformConcrete
from .transform_executor import TransformExecutor, TransformState, TransformResult
from ..transforms import Transform
from ..utils import get_logger_shortcuts
from ..workflow import InstructionFactories as instr
from ..workflow.stack import HandlerStatus

class Transformator:
    def __init__(self,
            storage,
            transformations=[],
            logger=None
        ):

        self.storage = storage
        self.logger = logger
        self.log = get_logger_shortcuts(logger)

        # Hack when lists/tuples are included
        new_transformations = []
        for x in transformations:
            if isinstance(x, tuple) or isinstance(x, list):
                for tr in x:
                    new_transformations.append(tr)
            else:
                new_transformations.append(x)
        transformations = new_transformations

        for t in transformations:
            assert t.NAME != Transform.DEFAULTS['NAME'], \
                f"Attempting to construct Transformator from untitled transformation '{t.__name__}'"
        self.transformations = {t.NAME: t for t in transformations}
        self.graph = nx.DiGraph()

        for tr_name, tr_obj in self.transformations.items():
            assert not self.graph.has_node(tr_name), f"Transform '{tr_name}' is already registered"
            for source in tr_obj.SOURCE_ITEMS:
                self.graph.add_edge(source, tr_name, transform=tr_obj)
                self.graph.nodes[source]['is_item'] = True
            for target in tr_obj.TARGET_ITEMS:
                self.graph.add_edge(tr_name, target, transform=tr_obj)
                self.graph.nodes[target]['is_item'] = True
            for item in tr_obj.NOTE_ITEMS:
                self.graph.add_node(item, is_item=True)
            self.graph.nodes[tr_name]['is_item'] = False
            self.graph.nodes[tr_name]['transform'] = tr_obj

        for item_node in self.graph.nodes:
            if not self.graph.nodes[item_node]['is_item']:
                continue
            
            assert item_node in self.storage, f"Unregistered item '{item_node}'"
            # assert self.graph.in_degree(item_node) + self.graph.out_degree(item_node) > 0, \
            #     f"Item '{item_node}' is not involved in any transforms"

            self.graph.nodes[item_node]['item'] = getattr(storage, item_node)
        
        for item_name, _ in self.storage:
            if not self.graph.has_node(item_name):
                self.log.error(f"Item '{item_name}' is not involved in any transforms")
            
    def _find_path_to(self, target_name: str) -> list:
        assert self.graph.has_node(target_name), f"Transform graph does not contain the target node '{target_name}'"

        # Isolate connected component containing the target
        active_graph = nx.Graph()
        active_graph.add_edges_from(self.graph.edges)
        active_graph.add_nodes_from(self.graph.nodes)
        for comp in nx.connected_components(active_graph):
            if target_name in comp:
                main_component_set = comp
                break
        
        ag = AccessGraph(self.graph, main_component_set, logger=self.logger)

        found_seqs = []
        starting_items = []
        for seq in ag.paths_iterate(target_name):
            found_seqs.append(seq)
            starting_items.append(ag.get_all_ends(seq))
            break
        
        assert len(found_seqs) != 0, f"No route to target '{target_name}' was found"
        assert not len(found_seqs) > 1, f"Several paths to the target '{target_name}' were found. {repr(found_seqs)}"

        found_seq = found_seqs[0]
        
        starting_items = starting_items[0]
        startitems_message = "Initial DataItems: " + ", ".join(starting_items)
        seq_message = "Sequence of transforms: " + " -> ".join(found_seq)
        self.log.info("\n\n{sep}\n{startitems_message}\n{seq_message}\n{sep}" \
            .format(
                startitems_message=startitems_message,
                seq_message=seq_message,
                sep=''.join(['-']*len(seq_message))
            )
        )
        return found_seq, starting_items
    
    def contains_dataitem(self, item_name: str) -> bool:
        return self.graph.has_node(item_name) and self.graph.nodes[item_name]['is_item']

    def plan_transformation(self,
            stack_frame,
            transformator_name: str,
            target: str,
            sources: list=None,
            forward: dict={}
        ) -> None:

        if sources is not None:
            raise NotImplementedError
        transform_path, start_items = self._find_path_to(target)

        # TODO Do some checks with startitems

        instructions = []
        for transform_name in transform_path:
            instruction = instr.prepare_transform(
                transformator=transformator_name,
                transform=transform_name,
                path=transform_path,
                forward=forward
            )
            instruction.info.activates = []
            instruction.info.activated_by = []
            instructions.append(instruction)
        for prev_tr, next_tr in zip(instructions, instructions[1:]):
            prev_tr.info.activates.append(next_tr.info.name)
            next_tr.info.activated_by.append(prev_tr.info.name)
        stack_frame.include_instructions(instructions)

    def prepare_transform(self,
            stack_frame,
            transformator_name: str,
            transform_name: str,
            transform_path: list,
            forward: dict={}
        ) -> None:

        transform_inst = TransformConcrete(
            transformator=self,
            transform_name=transform_name,
            logger=self.logger
        )
        transform_inst.prepare_instance(forward=forward)
        
        instructions = []
        for key_combination in transform_inst.get_key_combinations():
            # By default, entry point is the 'exec'-method for all transform
            instruction = instr.execute_transform(
                transformator=transformator_name,
                transform=transform_name,
                unmerged_keys=key_combination,
                forward=forward
            )
            instruction.info.activates = []
            instruction.info.activated_by = []
            instructions.append(instruction)
        
        stack_frame.include_instructions(instructions)
        stack_frame.info.transform_instance = transform_inst
    
    def execute_transform(self,
            stack_frame,
            instruction,
            transformator_name: str,
            transform_name: str,
            unmerged_keys: dict,
            forward: dict={}
        ) -> HandlerStatus:

        if 'keys_control' in forward and not forward['keys_control'](unmerged_keys):
            return HandlerStatus.DONE # Skip transform if user requested so
        
        if 'transform_executor' not in instruction.info:
            instruction.info.transform_executor = TransformExecutor(
                transform_instance=stack_frame.info.transform_instance,
                transformator_name=transformator_name,
                transform_name=transform_name,
                logger=self.logger
            )
        transform_executor = instruction.info.transform_executor

        result: TransformResult = transform_executor.execute(
            unmerged_keys=unmerged_keys,
            forward=forward,
        )

        if result == TransformResult.FINISHED or result == TransformResult.FAILED:
            return HandlerStatus.DONE
        elif result == TransformResult.REPEAT:
            return HandlerStatus.REPEAT
        elif result == TransformResult.LATER:
            return HandlerStatus.LATER
        else:
            raise ValueError(f"Cannot process transform result '{result}'")



import os, shutil
import random
import subprocess
import shlex

try:
    # Do this for all relative imports
    from ..utils import get_logger_shortcuts
except:
    pass


class RandomWord:
    vowels = 'AEIOU'
    consonants = 'BCDFGHJKLMNPQRSTVWXYZ'

    @staticmethod
    def _generate_word():
        word = ''
        for letter in RandomWord._random_pattern():
            if letter == 'V':
                word += random.choice(RandomWord.vowels)
            elif letter == 'C':
                word += random.choice(RandomWord.consonants)
        return word

    @staticmethod
    def _random_pattern():
        elements = ['CV', 'CV', 'VC', 'VCV', 'CVC']
        pattern = ""
        for i in range(random.randint(2, 4)):
            pattern += random.choice(elements)
        return pattern

    @staticmethod
    def get_random_word(reserved_words=[]):
        new_word = RandomWord._generate_word()
        while new_word in reserved_words:
            new_word = RandomWord._generate_word()
        return new_word


class ThreadManager:
    def __init__(self, wd: str, maxproc: int, logger=None) -> None:
        self.logger = logger
        self.log = get_logger_shortcuts(logger)

        self.wd = os.path.abspath(wd)
        if os.path.isdir(self.wd):
            shutil.rmtree(self.wd)
        os.mkdir(self.wd)

        self.maxproc = maxproc
        self.occupied_procs = 0
        self.pending_tasks = {}
        self.running_tasks = {}
        self.finished_tasks = {}
        self.registered_calc_ids = []

    def id_to_calcdir(self, id):
        return os.path.join(self.wd, id)
    
    def prepare_new_job(self, use_calcdir=True):
        job_dir = None
        while job_dir is None or os.path.isdir(job_dir):
            job_id = RandomWord.get_random_word()
            job_dir = self.id_to_calcdir(job_id)

        self.registered_calc_ids.append(job_id)
        if use_calcdir:
            job_dir = self.id_to_calcdir(job_id)
            assert not os.path.isdir(job_dir)
            os.mkdir(job_dir)
            return job_id, job_dir
        else:
            return job_id

    def check_queue(self):
        # Process finished tasks
        finished_ids = []
        for id, task in self.running_tasks.items():
            process = task['process']
            if process.poll() is None:
                continue

            # If finished:
            self.occupied_procs -= task['nproc']

            code = process.returncode
            if code == 0:
                self.log.info(f"Calc '{task['command']}' has finished")
            else:
                self.log.error(f"Calc '{task['command']}' has finished with ERROR")

            finished_ids.append({
                'id': id,
                'code': code,
            })
        
        for data in finished_ids:
            id, code = data['id'], data['code']
            del self.running_tasks[id]
            self.finished_tasks[id] = code
        
        # Start some pending tasks
        started_ids = []
        for id, task in self.pending_tasks.items():
            if task['nproc'] + self.occupied_procs > self.maxproc:
                continue

            if task['wd'] is not None:
                main_wd = os.getcwd()
                # raise Exception(repr(task['wd']))
                os.chdir(task['wd'])
            process = subprocess.Popen(shlex.split(task['command']))
            if task['wd'] is not None:
                os.chdir(main_wd)

            self.occupied_procs += task['nproc']

            self.running_tasks[id] = {
                'nproc': task['nproc'],
                'command': task['command'],
                'process': process
            }
            started_ids.append(id)

        for id in started_ids:
            del self.pending_tasks[id]

    @staticmethod
    def function_call_line(function_name, script_path, args):
        script_name = os.path.basename(script_path).replace('.py', '')
        return 'python -c "from {} import {}; {}(*{})"'.format(
            script_name,
            function_name,
            function_name,
            repr(args)#.replace("'", '\\"')
        )

    def run_nonblocking(self, id, command, nproc=1, wd=None):
        assert nproc <= self.maxproc, f"nproc(={nproc}) must be less than max allowed(={self.maxproc})"
        
        self.pending_tasks[id] = {
            'id': id,
            'command': command,
            'wd': wd,
            'nproc': nproc,
        }
            
        self.check_queue()
    
    def is_finished(self, id):
        self.check_queue()

        assert id in self.registered_calc_ids, f"CalcID={id} is not registered"
        assert id in self.pending_tasks or id in self.running_tasks or id in self.finished_tasks
        return id in self.finished_tasks

    def finalize_task(self, id, state='finished'):
        if state == 'finished':
            del self.finished_tasks[id]
        elif state == 'registered':
            pass
        else:
            raise RuntimeError(f"Job state '{state}' is not implemented")
        del self.registered_calc_ids[self.registered_calc_ids.index(id)]



from .storage import DataStorage
from .dataitems import ObjectItem, FileItem, DirectoryItem
from .planner import Transformator
from .transforms import Transform
from .workflow import Workflow, InstructionFactories
from .sidebar_ui import execute_with_ui
from .planner import TransformStateFactories
from .worker import run_worker
from .threadmanager import ThreadManager
from .planner.transform_utils import TransformState
